/*------------------------------------------------------------------------
 Alternating direction augmented Lagrangian method 

 Common routines
  
 Auther: Zaiwen Wen
 Date:   08-01-2009 
------------------------------------------------------------------------*/

#include "AltSDP_util.h"
#include "mex.h"


/*-----------------------------------------------------------------------
 * X is dense matrix, XX is dense vector
 * C is a sparse matrix, CX is a column of C
 *
 * Sparse matrix vector operation:
 * X*CX
 * CX'*XX
-----------------------------------------------------------------------
 

Sparse matrices have a different storage convention from that of full matrices
in MATLAB. The parameters pr and pi are still arrays of double-precision numbers,
but these arrays contain only nonzero data elements. There are three additional
parameters: nzmax, ir, and jc.
 
nzmax is an integer that contains the length of ir, pr, and, if it exists, pi.
It is the maximum possible number of nonzero elements in the sparse matrix.
 
ir points to an integer array of length nzmax containing the row indices of
the corresponding elements in pr and pi.
 
jc points to an integer array of length n+1, where n is the number of columns
in the sparse matrix. The jc array contains column index information. If the
jth column of the sparse matrix has any nonzero elements, jc[j] is the index
in ir and pr (and pi if it exists) of the first nonzero element in the jth column,
and jc[j+1] - 1 is the index of the last nonzero element in that column.
For the jth column of the sparse matrix, jc[j] is the total number of nonzero
elements in all preceding columns. The last element of the jc array, jc[n],
is equal to nnz, the number of nonzero elements in the entire sparse matrix.
If nnz is less than nzmax, more nonzero entries can be inserted into the array
without allocating additional storage.
 */

#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif

/* print a matrix */
void printVec(int m,  double *X)
{
    int di;
    mexPrintf("\n");
    for(di=0; di<m; di++)
        mexPrintf("%3.2e \t", X[di]);
    mexPrintf("\n");
}

/* print a matrix */
void printMat(int m, int n,  double *X)
{
    int di, dj;
    mexPrintf("\n");
    for(di=0; di<m; di++)
    {
        for(dj=0; dj<n; dj++)
            mexPrintf("%3.2e \t", *(X+dj*m+di));
        mexPrintf("\n");
    }
}

void printMatInt(int m, int n,  int *X)
{
    int di, dj;
    mexPrintf("\n");
    for(di=0; di<m; di++)
    {
        for(dj=0; dj<n; dj++)
            mexPrintf("%d \t", *(X+dj*m+di));
        mexPrintf("\n");
    }
}


double sign(double a)
{
    double EPS = 1e-30;
    if(a>=EPS) return 1.;
    else if (a<=-EPS) return -1.;
    else return 0.;
}

/* create a zero vector */
void zeros(int n,  double *X)
{
    int di;
    for( di=0; di <n; di++)
        X[di] = 0.0;
}

/* create a zero vector */
void ones(int n,  double *X)
{
    int di;
    for( di=0; di <n; di++)
        X[di] = 1.0;
}

double sumSYMX(int n,  double *X)
{
    int di, dj;
    double sum=0;
    for( di=0; di <n; di++)
    {
        sum += X[di*n+di];
        for (dj = di+1; dj<n; dj++)
            sum += 2*X[dj*n+di];
    }
    return sum;
}

/* copy W to Wold*/
void copySYM(int n, double *W, double *Wold)
{
    int di, dj, idx;
    for( di=0; di <n; di++)
    {
        for (dj = di; dj<n; dj++)
        {
            idx = dj*n+di; Wold[idx] =W[idx];
        }
    }
}

/* scale W = alpha*W*/
void scalSYM(int n, double *W, double alpha)
{
    int di, dj, idx;
    for( di=0; di <n; di++)
    {
        for (dj = di; dj<n; dj++)
        {
            idx = dj*n+di; W[idx] =W[idx]*alpha;
        }
    }
}

/* F-norm of (W-Wold) */
double nrmdiffSYM(int n, double *Wold, double *W)
{
    double dtmp, diff = 0.;
    int di, dj, idx;
    for( di=0; di <n; di++)
    {
        idx = di*n+di;  dtmp = Wold[idx] - W[idx]; diff += (dtmp*dtmp);
        for (dj = di+1; dj<n; dj++)
        {
            idx = dj*n+di;
            dtmp = Wold[idx] - W[idx]; diff += 2.0*(dtmp*dtmp);
        }
    }
    diff = sqrt(diff);
    return diff;
}


double nrmSYM(int n, double *W)
{
    double dtmp, diff = 0.;
    int di, dj, idx;
    for( di=0; di <n; di++)
    {
        idx = di*n+di;  dtmp = W[idx]; diff += (dtmp*dtmp);
        for (dj = di+1; dj<n; dj++)
        {
            idx = dj*n+di;
            dtmp = W[idx]; diff += 2.0*(dtmp*dtmp);
        }
    }
    diff = sqrt(diff);
    return diff;
}

/* if plus = 0, X = Y-Z
    if plus = 1, X = Y+Z
 */
void diffSYM(int n, double *X, double *Y, double *Z, int plus)
{
    int di, dj, idx;
    if(plus==0)
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;     X[idx] = Y[idx] - Z[idx];
            }
        }
    }
    else if (plus==1)
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;     X[idx] = Y[idx] + Z[idx];
            }
        }
    }
}


/*  X = a*Y+b*Z  */
void addSYM(int n, double *X, double a, double *Y, double b, double *Z)
{
    int di, dj, idx;

    for( di=0; di <n; di++)
    {
        for (dj = di; dj<n; dj++)
        {
            idx = dj*n+di;     X[idx] = a*Y[idx] + b*Z[idx];
        }
    }
 
}

double ddotSYM(int n,  double *X, double *Z)
{
    int di, dj, dl;
    double sum=0.0;
    for( di=0; di <n; di++)
    {
        sum += X[di*n+di]*Z[di*n+di];
        for (dj = di+1; dj<n; dj++)
            sum += 2.0*X[dj*n+di]*Z[dj*n+di];
    }
    return sum;
}


double matrix_norm(int n, double *a, int type)
/*                  PARAMETERS(input):                                */
/* int n;           Dimension of matrix.                              */
/* double *a;       Array storing matrix elements                     */
/*                  a   = *(a+i*n+j), i = 0(1)n-1, j = 0(1)n-1.       */
/*                   i,j                                              */
/* int type;        Type of matrix norm to be computed.               */
/*                  type = 1, 1- norm (maximum absolute column sum).  */
/*                       = 2, maximum- norm (maximum absolute row     */
/*                            sum).                                   */
/*                       = 3, E- norm (Frobenius norm).               */
/* The computed norm is returned through the function routine name.   */
/*                                                                    */
/**********************************************************************/
{
    
    int i, j;
    double value, cof;
    
    value = 0.;
    if (type == 1)
    {
        for (j = 0; j < n; j++)
        {
            cof = 0.;
            for (i = 0; i < n; i++)
                cof += fabs(*(a + i*n + j));
        }
        if (cof > value)
            value = cof;
    }
    else if (type == 3)
    {
        for (i = 0; i < n; i++)
            for (j = 0; j < n; j++)
                value += *(a + i*n + j) * *(a + i*n + j);
        value = sqrt(value);
    }
    else if(type == 2)
    {
        for (i = 0; i < n; i++)
        {
            cof = 0.;
            for (j = 0; j < n; j++)
                cof += fabs(*(a + i*n + j));
        }
        if (cof > value)
            value = cof;
    }
    else
    {
        mexErrMsgTxt("Function matrix_norm(), wrong value for type");
         /*printf("\n Function matrix_norm(), wrong value for type.");
        exit(); */
    }
    
    return(value); /* End of matrix_norm().                         */
}


/* eigenvector has been scaled: V(:,1)*sqrt(D(1)), ..., V(:,nev)*sqrt(D(nev)) 
void eigdsyevd(int n, double *W,  double *D, int *nev, int compEigS, int pos, double *work, int *iwork, int *ifail) */

/* Be careful!,  W should be  C - A'(y) - mu*X; */
void AltSDPPROJFEIG(int n, int nsq, double mu, double *S, double *X, double rho, double *XX, int* compEigS, double *W, double *WW, 
             double *D, int* nev, double *work, int *iwork)
{
    
    int  info,  neigp, neign;
    int  lwork = 1+6*n+2*n*n, liwork = 3+5*n;
    char  *jobz="V", *uplo="U";
    
    int di, dj, idx;
    
    double alpha, beta;
    int inc=1, incy;
    char *NTRANS = "N", *TTRANS = "T";    
    
    /*
    mexPrintf("matrix W: \n");
    printMat(n, n, W);
     */
    
    dsyevd_(jobz, uplo, &n, WW, &n, D,  work, &lwork, iwork, &liwork, &info);

    /*
    mexPrintf("eigen value: \n");
    printVec(n, D);
    */
    
    /*
    mexPrintf("eigen vector: \n");
        printMat(n, nev,V);
     */
    
    /* V(:,1)*sqrt(D(1)), ..., V(:,nev)*sqrt(D(nev)) */
    neign = 0;  neigp = 0;  
    for(di=0; di<n; di++)
    {
        if( D[di] < -1e-16 )
            neign ++;
        else if ( D[di] > 1e-16 )
            neigp ++;
    }
    
   /*  mexPrintf("n: %d, neign: %d, neigp: %d \n", n, neign, neigp); */
    /* negative eigenvalue corresponding to X
        positive eigenvalue corresponding to S
     */
    
    if (neign < neigp)  
    {
        *compEigS = 0;         *nev = neign;
        if(neign>0)
        {
            /* scale eigen vector */
            for(di = 0; di<neign; di++)
            {
                alpha = sqrt(-D[di]);
                dscal(&n, &alpha, WW+di*n, &inc);
            }
            alpha = 1; beta = 0.; dsyrk(uplo, NTRANS, &n, nev, &alpha, WW, &n, &beta, XX, &n);
            
        }
        else
            zeros(nsq, XX);
        
            /* S = (S+W); X = X/mu; Be careful: W = -W in this case*/
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                S[idx] = (XX[idx] + W[idx]);  X[idx] = (1-rho)*X[idx] + rho*XX[idx]/mu;
            }
        }
    }
    else
    {
        *compEigS = 1;       *nev = neigp;
        if(neigp>0)
        {
            for(di = n-neigp; di<n; di++)
            {
                alpha = sqrt(D[di]);
                dscal(&n, &alpha, WW+di*n, &inc);
            }
            alpha = 1; beta = 0.; dsyrk(uplo, NTRANS, &n, nev, &alpha, WW+(n-neigp)*n, &n, &beta, S, &n);
             
        }
        else
            zeros(nsq, S);
        
            /* X = (S-W)/mu; */
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                X[idx] = (1-rho)*X[idx] + (S[idx] - W[idx])*rho/mu;
            }
        }        
        
    }
    
}


/* eigenvector has been scaled: V(:,1)*sqrt(D(1)), ..., V(:,nev)*sqrt(D(nev)) */
void eigdsyevx(int n, double *W, double VL, double VU, double abstol,
        double *V, double *D, int *nev,int lwork, double *work, int *iwork, int *ifail)
{
    
    int   IL, IU, info, di, inc = 1;
    char  *jobz="V", *uplo="U",  *range="V";
    double alpha;
    
    /*
    mexPrintf("matrix W: \n");
    printMat(n, n, W);
     */
    
    dsyevx_(jobz, range, uplo, &n, W, &n, &VL, &VU, &IL, &IU,
    &abstol, nev, D, V, &n, work, &lwork, iwork, ifail, &info);
    /*
    mexPrintf("info: %d, nev: %d, VL: %e, VU: %e \n", info, *nev, VL, VU);
     
    mexPrintf("eigen value: \n");
    printVec(*nev, D);
     */
    
    /*
    mexPrintf("eigen vector: \n");
        printMat(n, nev,V);
     */
    
    /* V(:,1)*sqrt(D(1)), ..., V(:,nev)*sqrt(D(nev)) */
    for(di = 0; di<*nev; di++)
    {
        alpha = sqrt(D[di]);
        dscal(&n, &alpha, V+di*n, &inc);
    }
}


void AltSDPPROJ(int n, int nsq, double mu, double *S, double *X, double rho, double *XX, int* compEigS, double *W, double *WW, 
            double VL, double VU, double abstol, double *V, double *D, int* nev, int lwork, double *work, int *iwork, int *ifail )
{
    int di, dj, idx;
    
    double alpha, beta;
    int inc=1, incy;
    char *NTRANS = "N", *TTRANS = "T", *uplo="U";;
    
    eigdsyevx(n, WW, VL, VU, abstol, V, D, nev, lwork, work, iwork, ifail);

    if(*compEigS==1)
    {
        if(*nev>0)
        {
            alpha = 1; beta = 0.; dsyrk(uplo, NTRANS, &n, nev, &alpha, V, &n, &beta, S, &n);
        }
        else
            zeros(nsq,S);
        
            /* X = (S-W)/mu; */
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                X[idx] = (1-rho)*X[idx] + (S[idx] - W[idx])*rho/mu;
            }
        }
        
          
        
        if (*nev>n/2.0)
            *compEigS = 0;
    }
    else if (*compEigS==0)
    {
        if(*nev>0)
        {
            alpha = 1; beta = 0.; dsyrk(uplo, NTRANS, &n, nev, &alpha, V, &n, &beta, XX, &n);
        }
        else
            zeros(nsq, XX);
            /* S = (S+W); X = X/mu; Be careful: W = -W in this case*/
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                S[idx] = (XX[idx] - W[idx]);  X[idx] = (1-rho)*X[idx] + rho*XX[idx]/mu;
            }
        }
        
        if (*nev>n/2.0)
            *compEigS = 1;
    }
    
}


/* Input: V is the scaled eigenvectors or the solution of min || V-W||^2 ( min || V+W||^2)     
 */

/*
void AltSDPPROJLBFGS(int n,  double mu, double *S, double *X, double rho, double* XX, int* compEigS, double *W, double *WW, 
              double *V, int* nnev,  int mstore, double eps, double xtol, int mxitr, int record, double *workLB, double *work)
{
    int di, dj, idx;
    double f, nrmW;
    
    double alpha, beta;
    int inc=1, incy;
    char *NTRANS = "N", *TTRANS = "T", *uplo="U", *side = "L";

    
    int diagco, icall, iflag, iprint[2];
    int nvar, nev;
    
    nev = *nnev;
    
    nvar = n*nev;
    diagco = 0;
	iprint[0]= -2;
	iprint[1]=0;
	iflag = 0;
	icall = 0;
 
     nrmW = nrmSYM(n,W);
     
    while(1)
    {
        if(*compEigS==1)
        {
            alpha = 1.0; beta = 0.; dsyrk(uplo, NTRANS, &n, &nev, &alpha, V, &n, &beta, S, &n);
            diffSYM(n, XX, S, W, 0);
            f = nrmSYM(n,XX);  f = 0.5*f*f/nrmW;
            alpha = 1.0/nrmW; dsymm (side, uplo, &n, &nev, &alpha, XX, &n, V, &n, &beta, WW, &n);
        }
        else if (*compEigS==0)
        {
            alpha = 1.0; beta = 0.; dsyrk(uplo, NTRANS, &n, &nev, &alpha, V, &n, &beta, XX, &n);
   
            diffSYM(n, S, XX, W, 0);
            f = nrmSYM(n,S);  f = 0.5*f*f;
            alpha = 1.0/nrmW; dsymm (side, uplo, &n, &nev, &alpha, S, &n, V, &n, &beta, WW, &n);
        }
        
        if(record >= 2)
            mexPrintf("icall: %d, f: %3.2e,  nrmg: %3.2e\n", icall, f, dnrm2(&nvar, WW, &inc) );

        lbfgs_(&nvar, &mstore, V, &f, WW, &diagco, work, iprint, &eps, &xtol, workLB, &iflag);

        
        if(iflag<=0 || icall>=mxitr)
            break;
        ++icall;
	}
    
    addSYM(n, X, 1-rho, X, rho/mu, XX);
}

*/










