/***********************************************************************
 * mexARPACK.c : C mex file
 *
 * mex -O -largeArrayDims -lmwlapack -lmwblas mexeig.c
 *
 * [V,D,flag] = mexeig(A,k,sigma, opts);
 * c   sigma:
 * c          'LA' - compute the NEV largest (algebraic) eigenvalues.
 * c          'SA' - compute the NEV smallest (algebraic) eigenvalues.
 * c          'LM' - compute the NEV largest (in magnitude) eigenvalues.
 * c          'SM' - compute the NEV smallest (in magnitude) eigenvalues.
 * c          'BE' - compute NEV eigenvalues, half from each end of the
 * c                 spectrum.  When NEV is odd, compute one more from the
 * c                 high end than from the low end.
 * c           (see remark 1 below)
 * c
 *
 * If k == n, do NOT use ARPACK
 *
 * Zaiwen Wen
 * August 2009
 ***********************************************************************/

#include <math.h>
#include <mex.h>
#include <matrix.h>
#include <string.h> /* needed for memcpy() */

#if !defined(MX_API_VER) || ( MX_API_VER < 0x07030000 )
typedef int mwIndex;
typedef int mwSize;
#endif

#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif

/* print a matrix */
void printVec(mwIndex m,  double *X)
{
    mwIndex di;
    mexPrintf("\n");
    for(di=0; di<m; di++)
        mexPrintf("%3.2e \t", X[di]);
    mexPrintf("\n");
}

/* print a matrix */
void printMat(mwIndex m, mwIndex n,  double *X)
{
    mwIndex di, dj;
    mexPrintf("\n");
    for(di=0; di<m; di++)
    {
        for(dj=0; dj<n; dj++)
            mexPrintf("%3.2e \t", *(X+dj*m+di));
        mexPrintf("\n");
    }
}

/**********************************************************
 *
 ***********************************************************/
void mexFunction(
    int nlhs,   mxArray  *plhs[],
    int nrhs,   const mxArray  *prhs[] )
{    
    double   *A, *V, *D, *evals, *evecs, *tmp, tol, *flag=0;
    
    int   m, n, nev, ncv;
    int      maxit, disp;
    char     *bmat="I",  *all = "A", sigma[2];
    
    double   *workl, *workd, *resid;
    int *select, iparam[11], ipntr[11], nconv;
    int ldv, ido, lworkl, info, ierr;
    int mode, ishfts, rvec;
    double zero;

    int i, nummatvec;
    double minevalest;
    int tempint;

    int inc = 1, incy =1;
    double alpha, beta;
    char *TRANS = "N";

/* CHECK FOR PROPER NUMBER OF ARGUMENTS */
    
    if (nrhs > 4){
        mexErrMsgTxt("mexeig: requires at most 4 input arguments."); }
    if (nlhs > 3){
        mexErrMsgTxt("mexeig: requires at most 3 output argument."); }
    
/* CHECK THE DIMENSIONS */
    
    m = mxGetM(prhs[0]);
    n = mxGetN(prhs[0]);
    if (m != n) {
        mexErrMsgTxt("mexeig: matrix must be square."); }
    if (mxIsSparse(prhs[0])) {
        mexErrMsgTxt("mexeig: sparse matrix not allowed."); }
    A = mxGetPr(prhs[0]);
    
    nev = (mwSize)*mxGetPr(prhs[1]);
    mxGetString(prhs[2], sigma, 3);
    
    /*Get fields from opts */
    maxit = 100; disp = 0; tol = 1e-8;
    tmp = mxGetPr(mxGetField(prhs[3], 0, "maxit"));  maxit = (int)*tmp;
    tmp = mxGetPr(mxGetField(prhs[3], 0, "disp"));   disp  = (int)*tmp;
    tmp = mxGetPr(mxGetField(prhs[3], 0, "tol"));    tol   = *tmp;
    
    plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
    flag = mxGetPr(plhs[2]);    *flag = 0;
    
    /***** Do the computations in a subroutine *****/
    ldv = n;
    ncv = MIN(MAX(2*nev,20),n);
    zero = 0.0;
    lworkl = ncv*(ncv+8);
    info = 0;
    ido = 0;
    ishfts = 1;
    mode = 1;
    iparam[1-1] = ishfts;
    iparam[3-1] = maxit;
    iparam[7-1] = mode;
    rvec = 1;

    nummatvec = 0;
    
    V       = (double*)mxCalloc(ldv*ncv, sizeof(double));
    workl   = (double*)mxCalloc(ncv*(ncv+8), sizeof(double));
    workd   = (double*)mxCalloc(3*n, sizeof(double));
    D       = (double*)mxCalloc(ncv*2, sizeof(double));
    resid   = (double*)mxCalloc(n, sizeof(double));
    select  = (int*)mxCalloc(ncv, sizeof(int));
    
    if(disp>=10)
    {
        printMat(n,n,A);
        mexPrintf("nev: %d, ncv: %d, bmat: %s, sigma: %s, maxit: %d, disp: %d, tol: %3.2e \n",
        nev, ncv, bmat, sigma, maxit, disp, tol);
    }

    
    do {
        dsaupd_(&ido, bmat, &n, sigma, &nev, &tol, resid, &ncv, V,
            &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
        if(ido == -1 || ido == 1) {
            alpha = 1; beta = 0; inc = 1;
            dgemv(TRANS, &n, &n, &alpha, A, &n, workd + ipntr[1-1] - 1, 
                    &inc, &beta, workd + ipntr[2-1] - 1, &inc);
            nummatvec = nummatvec + 1;
        }
        else if(info < 0) {
            if(disp > 0) mexPrintf("ARPACK error (info = %d).\n", info);
            *flag = info; 
            plhs[0] = mxCreateDoubleMatrix(1,0,mxREAL);
            plhs[1] = mxCreateDoubleMatrix(1,0,mxREAL);
            return;
        }
    } while(ido != 99);

    if(disp>=10)
        mexPrintf("dsaupd, info: %d, matvec: %d.\n", info, nummatvec);

    dseupd_(&rvec, all, select, D, V, &ldv, &sigma, bmat, &n, sigma, 
            &nev, &tol, resid, &ncv, V, &ldv, iparam, ipntr, workd, 
            workl, &lworkl, &ierr);
    
    if(ierr != 0) {
        if(disp > 0) mexPrintf("ARPACK error (ierr = %d).\n", ierr);
        *flag = ierr;     
        plhs[0] = mxCreateDoubleMatrix(1,0,mxREAL);
        plhs[1] = mxCreateDoubleMatrix(1,0,mxREAL);
        return;
    }

    nconv = iparam[5-1];
    
    if(disp>=10)
        mexPrintf("dseupd, ierr: %d, nconv: %d.\n", ierr, nconv);


    if(disp > 0 && nconv < nev) {
        mexPrintf("Warning: %d out of %d evals converged.", nconv, nev);
        if(nev == 1) mexPrintf(" Best estimate returned.\n");
        else mexPrintf("\n");
    }

    /***** create return argument *****/
    plhs[0] = mxCreateDoubleMatrix(n,nev,mxREAL);
    evecs = mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(nev,1,mxREAL);
    evals = mxGetPr(plhs[1]);
    
    dcopy(&nev, D, &inc, evals, &inc);
    tempint = nev*n;
    dcopy(&tempint, V, &inc, evecs, &inc);

    /*
    if(nev == 1 && *nconv == 0) {
        minevalest = 1.0e10;
        for(i = 1; i <= ncv; i++)
            minevalest = mymin(minevalest, workl[ipntr[6-1]-2+i]);
    evals[0] = minevalest;
    }*/

    if(nconv < nev) *flag = -1;
}
/**********************************************************/
