/*----------------------------------------------------------

 *[X, S, y, out] = mexAltSDP_FAP(n, m,k,p,q, full(X0), full(C),  IE, opts); 

 Alternating direction augmented Lagrangian method 
  on computing theta problem
 *
% min  <C, X>
% s.t. X_ij = -1/(k-1), for (i,j) in  U
%      X_ij >= -1/(k-1), for (i,j) in E\U
%      X_ii = 1, X psd.
%
%  Scale the constraints to be orthogonal


 input:
 * n:        SDP dimension
 * m:       number of constraints
 * k:        parameter k
 * p:       number of constraints in U
 * q:       number of constraints in E\U
 * X0:      initial solution
 * C:       objective coefficient matrix
 * IE:        Graph
 * opts:    options
 *
 *------ Generate IE---------
 * eqk is the set of X_ij in U and geq is the set of X_ij in E\U
    [row,col] = find(eqk);     E = [row, col];     IE = (col-1)*n+row;
    [row,col] = find(geq);     E = [E; [row, col]];    IE = [IE; (col-1)*n+row];
 
  output:
 * X:       primal solution
 * S:       dual variable
 * y:       Lagrangian multiplier for linear constraints
 * out:    information*

  
 Auther: Zaiwen Wen
 Date:   08-01-2009 

 ----------------------------------------------------------*/

#include "AltSDP_util.h"
#include "mex.h"



/* b=  [1, -sqrt(2)/(k-1) ] */
double Computedobj(int n, int m, int k, double *y)
{
    int di;
    double dobj = 0.0;
    for(di=0; di<n; di++)
        dobj  += y[di];
    for(di= n; di<m; di++)
        dobj -= (y[di]*sqrt(2.0)/(k-1));
    return dobj;
}

/* AX - b >=0  */
double ComputeNrmResi(int n, int p, int q, double *resi)
{
    int di;
    double nrm = 0.0, dtmp;
    for(di=0; di<n+p; di++)
    {
        dtmp = resi[di];  nrm += (dtmp*dtmp);
    }
    
    for(di= n+p+1; di<n+p+q; di++)
    {
        dtmp = MIN(resi[di], 0.0);  nrm += (dtmp*dtmp);
    }
    
    return sqrt(nrm);
}

/* AX = [diag(X), X(T)*sqrt(2);]; */
void ComputeAX(int n, int pq, double *X, unsigned int *IE, double *y)
{
    int di;
    for(di=0; di<n; di++)
        y[di] = X[di*n+di];
    
    for(di=0; di<pq; di++)
        y[di+n] = sqrt(2.0)*X[IE[di]];
    
}


/* AX = [diag(X), X(T)*sqrt(2);] - [1, -sqrt(2)/(k-1) ]*/
void ComputeAXresi(int n, int pq, int k, double *X, unsigned int *IE, double *y)
{
    int di;
    for(di=0; di<n; di++)
    {
        y[di] = X[di*n+di]-1.0;
    }
    
    for(di=0; di<pq; di++)
    {
        y[di+n] = sqrt(2.0)*(X[IE[di]] + 1.0/(k-1));
    }
    
}

/*
ATy = zeros(n,n);
ATy(IE) = y(n+1:m)/sqrt(2);
ATy = ATy + ATy'+ diag(y(1:n));
 */
/* Here, we only store the upper triangle of the matrix ATy*/
void ComputeATy(int n, int pq, double *y, unsigned int *IE, double *ATy)
{
    int di;
    for(di=0; di<n; di++)
        ATy[di*n+di] = y[di];
    
    for(di=0; di<pq; di++)
        ATy[IE[di]] = y[di+n]/sqrt(2.0);
}

void ComputeW(int n, double *C, double *ATy, double mu, double *X, double *W, double *WW, bool compEigS)
{
    unsigned int di, dj, idx;
    if(compEigS)
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                W[idx] = C[idx] -ATy[idx] - mu*X[idx]; WW[idx] = W[idx];
            }
        }
    }
    else
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                W[idx] = ATy[idx] + mu*X[idx] - C[idx];   WW[idx] = W[idx];
            }
        }
    }
    
}



/* The gateway routine */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  /* Input */
    double  *XX, *E, *C;
    int     m, n, k, p, q, pq;
    unsigned int *IE;
    
    /* fields for options */
    double  ftol, mu, rmu, mu_min, mu_max, rho , delta_mu_l, delta_mu_u;
    int     record, mxitr, min_mu_itr, max_mu_itr, max_itr_stag_lev1, max_itr_stag_lev2, max_itr_stag_lev3;
    
  /* Output */
    double *X, *S, *y;
    double  *flag;
    
    mxArray *fout[7];
    const char *fnames[] = {"pobj", "dobj", "gap", "pinf", "dinf",  "itr",  "exit"};
    const char *fexit[] = {"optimal", "exceed max iteration"};
    int   nfields = 7;
    double *ppobj, *pdobj, *pgap, *ppinf, *pdinf, *pitr;
    
 /* work variable */
    int   itr, di, dj, idx, itmu_pinf , itmu_dinf, itr_stag_lev1=0, itr_stag_lev2=0, itr_stag_obj=0, itr_stag = 0;
    unsigned int nsq;
    double *tmp, *AC, *ATy, *W, *resi, nrmC, nrmb, objscale, pinf, dinf, gap, dtmp;
    double ref_inf, rel_pobj, rel_dobj, pobj_old, dobj_old;

    int    compEigS, TcompEigS;
    
    /*variables for eigenvalue decomposition */
    double *WW, *V, *D, *work;
    int nev, lwork, *ifail, *iwork;
    
    
    double alpha, beta;
    int inc=1, incy;
    char *NTRANS = "N", *TTRANS = "T", *uplo="U";;
    
    
  /*  Check for proper number of arguments. */
  /* NOTE: You do not need an else statement when using
     mexErrMsgTxt within an if statement. It will never
     get to the else statement if mexErrMsgTxt is executed.
     (mexErrMsgTxt breaks you out of the MEX-file.)
   */
    if (nrhs != 9)
        mexErrMsgTxt("Nine inputs required.");
    if (nlhs != 4)
        mexErrMsgTxt("Four outputs required.");
    
    /* n, m, k, p, q,  X, C, IE,*/
    
    
    tmp = mxGetPr(prhs[0]); n = (int)*tmp;  nsq = n*n;
    tmp = mxGetPr(prhs[1]); m = (int)*tmp;
    tmp = mxGetPr(prhs[2]); k = (int)*tmp;
    tmp = mxGetPr(prhs[3]); p = (int)*tmp;
    tmp = mxGetPr(prhs[4]); q = (int)*tmp;      pq = p + q;
    XX = mxGetPr(prhs[5]);
    C = mxGetPr(prhs[6]);
    E = mxGetPr(prhs[7]);
    
  /* Create a pointer to the input matrix W, M, T. */
    if(mxIsSparse(prhs[5])) {mexErrMsgTxt("Matrix X must be a full matrix."); return;}
    if(mxIsSparse(prhs[6])) {mexErrMsgTxt("Matrix C must be a full matrix."); return;}
    
    
    /* mxGetM(prhs[2]); */
        /*
    if(record>=10)
    {
        printMat(n, n,XX);
    }
         */
    
    /*Get fields from opts */
    if (mxGetField(prhs[8], 0, "record") == NULL )         record = 0;
    else    {   tmp = mxGetPr(mxGetField(prhs[8], 0, "record"));    record      = (int)*tmp;}

    if (mxGetField(prhs[8], 0, "mxitr") == NULL )         mxitr = 100;
    else    {   tmp = mxGetPr(mxGetField(prhs[8], 0, "mxitr"));       mxitr       = (int)*tmp;    }
   
    if (mxGetField(prhs[8], 0, "min_mu_itr") == NULL )        min_mu_itr = 0;    
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "min_mu_itr"));    min_mu_itr   = (int)*tmp;    }    
    
    if (mxGetField(prhs[8], 0, "max_mu_itr") == NULL )        max_mu_itr = 0;    
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "max_mu_itr"));    max_mu_itr   = (int)*tmp;    }
    
    if (mxGetField(prhs[8], 0, "mu") == NULL )        mu = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "mu"));        mu          = *tmp;     }
        
    if (mxGetField(prhs[8], 0, "rmu") == NULL )        rmu = 0.1;  
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "rmu"));       rmu         = *tmp;    }
    
    if (mxGetField(prhs[8], 0, "mu_min") == NULL )        mu_min = 1;     
    else     {         tmp = mxGetPr(mxGetField(prhs[8], 0, "mu_min"));        mu_min         = *tmp;      }
    
    if (mxGetField(prhs[8], 0, "mu_max") == NULL )        mu_max = 100;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "mu_max"));        mu_max         = *tmp;      }

    if (mxGetField(prhs[8], 0, "ftol") == NULL )        ftol = 1e-4;    
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "ftol"));      ftol        = *tmp;    }
    
    if(mxGetField(prhs[8], 0, "rho") == NULL )        rho = 1.6;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "rho"));       rho     = *tmp;    }
    
    if (mxGetField(prhs[8], 0, "max_itr_stag_lev1") == NULL )        max_itr_stag_lev1 = 10;
   else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "max_itr_stag_lev1"));    max_itr_stag_lev1   = (int)*tmp;    }
    
    if (mxGetField(prhs[8], 0, "max_itr_stag_lev2") == NULL )        max_itr_stag_lev2 = 100;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "max_itr_stag_lev2"));    max_itr_stag_lev2   = (int)*tmp;    }
    
    if (mxGetField(prhs[8], 0, "max_itr_stag_lev3") == NULL )        max_itr_stag_lev3 = 1000;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "max_itr_stag_lev3"));    max_itr_stag_lev3   = (int)*tmp;    }  
    
    if (mxGetField(prhs[8], 0, "delta_mu_l") == NULL )         delta_mu_l = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "delta_mu_l"));       delta_mu_l =  *tmp;    }
    
    if (mxGetField(prhs[8], 0, "delta_mu_u") == NULL )         delta_mu_u = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[8], 0, "delta_mu_u"));       delta_mu_u = *tmp;    } 
    
    
    if(record>=10)
    {
        mexPrintf("mxitr: %d, ftol: %3.2e, mu: %3.2e, rmu: %3.2e, mu_min: %3.2e\n", mxitr, ftol,  mu, rmu, mu_min);
        mexPrintf("max_itr_stag_lev1: %d,  max_itr_stag_lev2: %d, max_itr_stag_lev3: %d\n", max_itr_stag_lev1, max_itr_stag_lev2, max_itr_stag_lev3);
    }
    
  /* Set the output pointer to the output variable */
    plhs[0] = mxCreateDoubleMatrix(n, n, mxREAL);      X  = mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(n, n, mxREAL);      S  = mxGetPr(plhs[1]);
    plhs[2] = mxCreateDoubleMatrix(m, 1, mxREAL);      y  = mxGetPr(plhs[2]);
    dcopy(&nsq, XX, &inc, X, &inc);
    
    
    plhs[3] = mxCreateStructMatrix(1, 1, nfields, fnames);
    /* char *fnames[] = {"pobj", "dobj", "nrmr", "itr",  "exit"}; */
    fout[0] = mxCreateDoubleMatrix(1,1, mxREAL);   ppobj = mxGetPr(fout[0]);
    fout[1] = mxCreateDoubleMatrix(1,1, mxREAL);   pdobj = mxGetPr(fout[1]);
    fout[2] = mxCreateDoubleMatrix(1,1, mxREAL);   pgap = mxGetPr(fout[2]);
    fout[3] = mxCreateDoubleMatrix(1,1, mxREAL);   ppinf = mxGetPr(fout[3]);
    fout[4] = mxCreateDoubleMatrix(1,1, mxREAL);   pdinf = mxGetPr(fout[4]);
    fout[5] = mxCreateDoubleMatrix(1,1, mxREAL);   pitr  = mxGetPr(fout[5]);
    fout[6] = mxCreateString(fexit[0]);
    
    
    ATy     = mxCalloc(nsq, sizeof(double));
    W       = mxCalloc(nsq, sizeof(double));
    AC      = mxCalloc(m, sizeof(double));
    resi    = mxCalloc(m, sizeof(double));
    IE      = mxCalloc(p+q, sizeof(unsigned int));
    
    
    /* allocate variable for dsyevx */
    lwork   = 8*n;
    WW      =  mxCalloc(nsq, sizeof(double));
    V       =  mxCalloc(nsq,sizeof(double));
    D       =  mxCalloc(n, sizeof(double));
    work    =  mxCalloc(lwork,sizeof(double));
    iwork   =  mxCalloc(5*n, sizeof(int));
    ifail   =  mxCalloc(n, sizeof(int));
    
    
    /*convert fortran index to c index*/
    for(di=0; di<p+q; di++)   IE[di] = (int)E[di]-1;
    /*
    if (record >= 10)
    {
        for(di=0; di<(m-1); di++)
            mexPrintf("E[%5d, %5d], idx: %5d\n", (int)E[di], (int)E[m-1+di], LE[di]);
    }
     
    ones(m,  y);
    ComputeATy(n, m, y, LE, ATy);
    printMat(n, n, ATy);
     
    ComputeAX(n, m, ATy, LE, y);
    printMat(m, 1, y);
     */
    
    /*scaling data*/
    /*
    nrmC = dnrm2(&nsq, C, &inc);  dtmp = 1 + nrmC;  alpha = 1/dtmp;
    dscal(&nsq, &alpha, C, &inc);
    objscale = dtmp;  nrmC = 1+nrmC/dtmp;
     */
    
    /* nrmC = 1 + dnrm2(&nsq, C, &inc);  objscale = 1; */
    nrmC = 1 + matrix_norm(n, C, 1);    objscale = 1;
     /*b=1 or -sqrt(2)/(k-1);  */
    nrmb = 1 + sqrt(n+(p+q)*2.0/((k-1)*(k-1)));
    
    /*only compute the upper triangle*/
    zeros(m,  y);   zeros(nsq, S);  zeros(nsq, ATy);
    *ppobj =  objscale* ddotSYM(n, X, C);    *pdobj = 0.0;
    gap = fabs(*ppobj-*pdobj)/(1.+fabs(*ppobj)+fabs(*pdobj));
    ComputeAX(n, pq, C, IE, AC);
    /* resi = AX - b*/
    ComputeAXresi(n, pq, k,  X, IE, resi);
    /*printVec(m, resi); */
    pinf = ComputeNrmResi(n, p, q, resi)/nrmb;  dinf = 1.;
    compEigS = true;
    
    itmu_pinf = 0;  itmu_dinf = 0;      itr_stag = 0;
    ref_inf = MAX(MAX(pinf, dinf), gap);
    
    
  /*  mexPrintf("n: %d, m: %d,  nrmb: %e\n", n, m,  nrmb); */
    
    if (record >= 1)
        mexPrintf("%4d   %3.2e   %+3.2e   %+3.2e   %+3.2e   %3.2e   %+3.2e  %4d (%d)\n",
        0, mu,  *ppobj,  *pdobj,  gap, pinf, dinf,  0, compEigS);
    
    for(itr = 1; itr<=mxitr; itr++)
    {
        pobj_old = *ppobj;  dobj_old = *pdobj;
         /* step 1: compute y
        y = (AC - ComputeAX(n, S, T) + mu*resi);  % compute dual y
          */
        
        ComputeAX(n, pq, S, IE, y);
        /*
        mexPrintf("AC:\n"); printVec(m, AC);
        mexPrintf("y:\n"); printVec(m, y);
        mexPrintf("resi:\n"); printVec(m, resi); */
        
        for(di=0; di<m; di++)
            y [di] = AC[di]-y[di]-mu*resi[di];
        
        for(di=n+p+1; di<m; di++)
        {
            if(y[di]<0) y[di] = 0.;
        }
        
        /* y [di] = MAX( y [di], 0); */
        
        /*
        alpha = -1;  daxpy(&m, &alpha, AC, &inc, y, &inc);  dscal(&m, &alpha, y, &inc);
        alpha = -mu; daxpy(&m, &alpha, resi, &inc, y, &inc);
         */
        /*
        mexPrintf("y:\n"); printVec(m, y);
         */
        
       /* step 2: compute S */
        ComputeATy(n, pq, y, IE, ATy);
        
        /* compute S by projection */
        /* W =  C - ATy - mu*X; */
        /*need an extra WW since it will be destoyed */
        ComputeW(n, C, ATy, mu, X, W, WW, compEigS);

        TcompEigS = compEigS;
        AltSDPPROJ(n, nsq, mu, S, X, rho, XX, &compEigS, W,  WW, 1e-12, 1e15, 1e-8, V, D, &nev, lwork, work, iwork, ifail); 
        /* AltSDPPROJ(n, nsq, mu, S, X, rho, XX, &compEigS, W,  WW, 1e-8, 1e10, 1e-6, V, D, &nev, lwork, work, iwork, ifail); */

        
        
    /* -------------------------check optimality-----------------*/
        *ppobj =  objscale* ddotSYM(n, X, C);
        *pdobj = objscale*Computedobj(n, m, k, y);
        gap = fabs(*ppobj-*pdobj)/(1+fabs(*ppobj)+fabs(*pdobj));
        ComputeAXresi(n, pq, k, X, IE, resi);
        pinf = ComputeNrmResi(n, p, q, resi)/nrmb;
        dinf = 0.; /* norm(ATy+S-C)/nrmC */
        for( di=0; di <n; di++)
        {
            idx = di*n+di;  dtmp = ATy[idx] + S[idx] - C[idx]; dinf += (dtmp*dtmp);
            for (dj = di+1; dj<n; dj++)
            {
                idx = dj*n+di;
                dtmp = ATy[idx] + S[idx] - C[idx]; dinf += 2*(dtmp*dtmp);
            }
        }
        dinf = sqrt(dinf)/nrmC;
        
        dtmp = MAX(MAX(pinf,dinf), gap);
        /* dtmp = MAX(pinf,dinf); */
        
        if (record >= 1)
            mexPrintf("%4d   %3.2e   %+3.2e   %+3.2e   %+3.2e   %3.2e   %+3.2e  %4d (%d) (%3.2e,   %+3.2e)  %4d  %3.2e\n",
            itr, mu,  *ppobj,  *pdobj,  gap, pinf, dinf,  nev, TcompEigS, ref_inf, dtmp, itr_stag, pinf/dinf);
        
        /* optimal solution */
        if( dtmp <= ftol )
            break;
        
        /* detection of stagnation */
        
        if (dtmp < ref_inf)
        {
            ref_inf = dtmp;  itr_stag = 0;
        }
        else
        {
            itr_stag = itr_stag + 1;
        }
        
        if ( ( itr_stag > max_itr_stag_lev1 &&  dtmp<= ftol*10) 
            || ( itr_stag > max_itr_stag_lev2 &&  dtmp<= ftol*100) 
            || ( itr_stag > max_itr_stag_lev3 && dtmp <= ftol*500) )
            break;
        
       
        /* update mu */


        dtmp = pinf/dinf;
        if (dtmp <= delta_mu_l)
        {
            itmu_pinf = itmu_pinf + 1;  itmu_dinf = 0;
            if(itmu_pinf > max_mu_itr)
            {
                mu = MAX(mu*rmu, mu_min); itmu_pinf = 0;
            }
        }
	
        else if (dtmp> delta_mu_u)
        {
            itmu_dinf = itmu_dinf + 1;  itmu_pinf = 0;
            if(itmu_dinf > max_mu_itr)
            {
                mu = MIN(mu/rmu, mu_max); itmu_dinf = 0;
            }
        }     


        /*
        if (pinf <= dinf)
        {
            itmu_pinf = itmu_pinf + 1;  itmu_dinf = 0;
            if(itmu_pinf > max_mu_itr)
            {
                mu = MAX(mu*rmu, mu_min); itmu_pinf = 0;
            }
        }
        else
        {
            itmu_dinf = itmu_dinf + 1;  itmu_pinf = 0;
            if(itmu_dinf > max_mu_itr)
            {
                mu = MIN(mu/rmu, mu_max); itmu_dinf = 0;
            }
        }
         */
        

        /*
        if (pinf < dinf )
            mu = MAX(mu*rmu, mu_min);
        */
        
    }
    
    
    *pgap = gap; *ppinf = pinf;  *pdinf = dinf; *pitr = itr;
    /*mexPrintf("itr: %d, mxitr: %d\n", itr, mxitr); */
    
    if(itr>=mxitr)
    {
        fout[6] = mxCreateString(fexit[1]);     *pitr = itr-1;
    }
    
    for(di=0; di < nfields; di++)
        mxSetField(plhs[3], 0, fnames[di], fout[di]);
    
    /* scaling back solution */
    for(di=0; di<n; di++)
        y[di] = objscale*y[di];    
    for(di=0; di<pq; di++)
        y[di+n] = objscale*sqrt(2.0)*y[di+n];
    
    for( di=0; di <n; di++)
    {
        idx = di*n+di;  S[idx] = objscale*S[idx];
        for (dj = 0; dj<di; dj++)
        {
            idx = dj*n+di; itr = di*n+dj;     X[idx] = X[itr];  
            S[itr] = objscale*S[itr];  S[idx] = S[itr];
        }
    }        
    
    
}
